#!/usr/local/bin/ruby

class JavaMethod
  attr :cumulusTime
  attr :selfTime

  def addCount(n)
    @count += n
  end
  def addCumulus(n)
    @cumulusTime += n
  end
  def addSelf(n)
    @selfTime += n
  end

  def initialize(name)
    @name = name
    @count = 0
    @cumulusTime = 0
    @selfTime = 0
  end

  def show()
    printf "%10d %10d %6d   %s\n", @cumulusTime, @selfTime, @count, @name
  end
end

class JavaMethodTable
  def initialize()
    @table = Hash.new()
  end

  def get(name)
    if @table.key?(name)
      method = @table[name]
    else
      method = JavaMethod.new(name)
      @table[name] = method
    end
    return method
  end

  def register(caller, callee, count, time)
#print "register called.\n"
#print " caller: ", caller, "\n"
#print " callee: ", callee, "\n"
#print " count: ", count, "\n"
#print " time: ", time, "\n"
    method = get(callee)
    method.addCount(count)
    method.addCumulus(time)
    method.addSelf(time)

    method = get(caller)
    method.addSelf(-time)
  end

  def show()
    print "cumulative   self      calls   name\n"
    print "      msec       msec\n"

    keys = @table.keys
    keys.sort! {|a, b| @table[a].selfTime < @table[b].selfTime}
    keys.each {|name| @table[name].show()}
  end
end


mtable = JavaMethodTable.new()

while fname = ARGV.shift()
  f = open(fname)
  while f.gets()
    next if not /^\s*[0-9]/
    chomp!()

    # <unknown caller>  <unknown_caller> ִ
    gsub! /<unknown\scaller>/, "<unknown_caller>"

    elems = split()
    count = elems[0]
    callee = elems[1]
    caller = elems[2]
    time = elems[3]

    mtable.register(caller, callee, count.to_i, time.to_i)
  end
end

mtable.show()
