/**
 * Sample code for reflection
 * in package NET.shudo.moba.reflect.
 * Compile:
 *   % javac MetaClassTest.java
 * and execute:
 *   % java MetaClassTest class_name [class_name ...]
 */

package NET.shudo.moba.examples;

import NET.shudo.moba.reflect.*;

class MetaClassTest {
  static final String COMMAND = "java MetaClassTest";

  private void start(String[] argv) {
    if (argv.length <= 0) {
      System.out.println("class name must be specified.");
      usage(COMMAND);  System.exit(1);
    }


    MetaClass mc = null;
    int fieldCount;
    int methodCount;
    int i, j;

    for (i = 0; i < argv.length; i++) {
      try {
	mc = MetaClass.get(argv[i]);
      }
      catch (ClassNotFoundException e) {
	e.printStackTrace();
	System.exit(1);
      }

      fieldCount = mc.fieldCount();
      methodCount = mc.methodCount();

      System.out.println("class: " + argv[i]);

      System.out.println("  #field  : " + fieldCount);
      System.out.println("  #method : " + methodCount);
      System.out.println("  #native : " + mc.nativeMethodCount());

      for (i = 0; i < methodCount; i++) {
	MetaMethod method = null;
	try { method = mc.method(i); }
	catch (FieldNotFoundException e) { continue; }

	System.out.println("    selector: " +
			method.name() + ", " + method.signature());
	System.out.println("    ID: " + method.hashCode());
      }

      System.out.println("  CP count: " + mc.constantpoolCount());

      for (i = 1; i < mc.constantpoolCount(); i++) {
	System.out.println("  CP type[" + i + "]: " + ConstantPool.type(mc.getClass(), i));
      }
    }
  }


  public static void usage(String command) {
    System.out.print("usage: ");
    System.out.print(command);
    System.out.println(" class_name ...");
  }

  public static void main(String[] argv) {
    new MetaClassTest().start(argv);
  }
}
