/*
 * Copyright (C) 1996,1997,1998,1999
 *	SHUDO Kazuyuki
 *
 * Permission to use, copy, modify and distribute this software
 * and its documentation for non-commercial purposes and without
 * fee is hereby granted provided this copyright notice.
 *
 * $Id$
 */

/* native methods for MetaMethod */

#include "NET_shudo_moba_reflect_MetaMethod.h"

#include "moba.h"


/*
 * ᥽åɤƤ
 */
JNIEXPORT void JNICALL Java_NET_shudo_moba_reflect_MetaMethod_invoke
  (JNIEnv *env, jobject _this, jobject obj, jobjectArray argAry) {
  jfieldID fid;
  jmethodID mid;
  jboolean isStatic;

  const char *signature;  jstring sigStr;
  char *retSig;	/* ֤ͤ signature */

  jsize argSize;
  jvalue *args;

  /* jmethodID  */
  fid = (*env)->GetFieldID(env, (*env)->GetObjectClass(env, _this),
				"methodID", "I");
  mid = (jmethodID)(*env)->GetIntField(env, _this, fid);
	/* attention!!! : Ѵ: jint -> jmethodID */

  /* static ɤ */
  {
    jint access;
    fid = (*env)->GetFieldID(env, (*env)->GetObjectClass(env, _this),
				"thisAccess", "I");
    access = (*env)->GetIntField(env, _this, fid);
    isStatic = ((access & 0x0008/* ACC_STATIC in tree.h */) ?
			JNI_TRUE : JNI_FALSE);
  }

  /* signature  */
  fid = (*env)->GetFieldID(env, (*env)->GetObjectClass(env, _this),
				"thisSig", "Ljava/lang/String;");
  sigStr = (jstring)(*env)->GetObjectField(env, _this, fid);
  signature = (*env)->GetStringUTFChars(env, sigStr, NULL);

  /* ֤ͤ signature  */
  retSig = signature;
  while (*retSig != ')')  retSig++;
  retSig++;

  /* Ѱդ */
  {
    jsize i;
    argSize = (*env)->GetArrayLength(env, argAry);
    args = (jvalue *)malloc(argSize * sizeof(jvalue));
    for (i = 0; i < argSize; i++)
      args[i].l = (*env)->GetObjectArrayElement(env, argAry, i);
  }

  /* ᥽åɸƤӽФ */
  if (isStatic) {	/* static ᥽å */
    jclass clazz;
    fid = (*env)->GetFieldID(env, (*env)->GetObjectClass(env, _this),
				"thisClass", "Ljava/lang/Class;");
    clazz = (jclass)(*env)->GetObjectField(env, _this, fid);

    switch (*retSig) {
    case 'L':
    case '[':
      (*env)->CallStaticObjectMethodA(env, clazz, mid, args);  break;
    case 'V':
      (*env)->CallStaticVoidMethodA(env, clazz, mid, args);  break;
    case 'Z':
      (*env)->CallStaticBooleanMethodA(env, clazz, mid, args);  break;
    case 'B':
      (*env)->CallStaticByteMethodA(env, clazz, mid, args);  break;
    case 'C':
      (*env)->CallStaticCharMethodA(env, clazz, mid, args);  break;
    case 'S':
      (*env)->CallStaticShortMethodA(env, clazz, mid, args);  break;
    case 'I':
      (*env)->CallStaticIntMethodA(env, clazz, mid, args);  break;
    case 'J':
      (*env)->CallStaticLongMethodA(env, clazz, mid, args);  break;
    case 'F':
      (*env)->CallStaticFloatMethodA(env, clazz, mid, args);  break;
    case 'D':
      (*env)->CallStaticDoubleMethodA(env, clazz, mid, args);  break;
    default:
#ifdef DEBUG
      printf("MetaMethod#invoke(): invalid signature: %s\n", signature);
      fflush(stdout);
#endif
      break;
    }
  }
  else {		/* not static ᥽å */
    switch (*retSig) {
    case 'L':
    case '[':
      (*env)->CallObjectMethodA(env, obj, mid, args);  break;
    case 'V':
      (*env)->CallVoidMethodA(env, obj, mid, args);  break;
    case 'Z':
      (*env)->CallBooleanMethodA(env, obj, mid, args);  break;
    case 'B':
      (*env)->CallByteMethodA(env, obj, mid, args);  break;
    case 'C':
      (*env)->CallCharMethodA(env, obj, mid, args);  break;
    case 'S':
      (*env)->CallShortMethodA(env, obj, mid, args);  break;
    case 'I':
      (*env)->CallIntMethodA(env, obj, mid, args);  break;
    case 'J':
      (*env)->CallLongMethodA(env, obj, mid, args);  break;
    case 'F':
      (*env)->CallFloatMethodA(env, obj, mid, args);  break;
    case 'D':
      (*env)->CallDoubleMethodA(env, obj, mid, args);  break;
    default:
#ifdef DEBUG
      printf("MetaMethod#invoke(): invalid signature: %s\n", signature);
      fflush(stdout);
#endif
      break;
    }
  }


  /*  */
  (*env)->ReleaseStringUTFChars(env, sigStr, signature);
  free(args);

#ifdef DEBUG
  if ((*env)->ExceptionOccurred(env)) {
    printf("MetaMethod#invoke():  An exception occurred.\n");
    (*env)->ExceptionDescribe(env);
    fflush(stdout);
  }
#endif

  return;
}


/*
 * method ID
 */
JNIEXPORT jint JNICALL Java_NET_shudo_moba_reflect_MetaMethod_methodID
  (JNIEnv *env, jclass _this, jclass clazz, jstring nameStr, jstring sigStr) {
  jmethodID mid;
  const char *name, *sig;

  sig = (*env)->GetStringUTFChars(env, sigStr, NULL);
  name = (*env)->GetStringUTFChars(env, nameStr, NULL);

#if 0
  printf("MetaMethod#methodID(): name `%s', signature `%s'\n", name, sig);
  fflush(stdout);
#endif
  mid = (*env)->GetMethodID(env, clazz, name, sig);
	// throws NoSuchMethodError
  if ((*env)->ExceptionOccurred(env)) {
    (*env)->ExceptionClear(env);
    mid = (*env)->GetStaticMethodID(env, clazz, name, sig);
	// throws NoSuchMethodError
  }

  (*env)->ReleaseStringUTFChars(env, sigStr, sig);
  (*env)->ReleaseStringUTFChars(env, nameStr, name);

  // attention!!! : sizeof(jint) == sizeof(jmethodID) 
  return (jint) mid;
}
