/*
 * Copyright (C) 1996,1997,1998,1999
 *	SHUDO Kazuyuki
 *
 * Permission to use, copy, modify and distribute this software
 * and its documentation for non-commercial purposes and without
 * fee is hereby granted provided this copyright notice.
 *
 * $Id$
 */

#include <stdio.h>	/* for NULL */
#include <string.h>	/* for memcpy() */
#include <stdarg.h>	/* for va_*() */

#include <jni.h>	/* for native methods */
#include <native.h>	/* for native methods */
#include <monitor.h>	/* for BINCLASS_{,UN}LOCK() */

#include "moba.h"
#include "opcodes.h"	/* for opc_* and opcode_length[] */


/*
 * Variables
 */
#define EXCEPTION_BUF_SIZE	100
static char exception_msg_buf[EXCEPTION_BUF_SIZE];


/*
 * constant pool item б륯饹롣
 */
extern ClassClass *cp2class(cp_item_type *cp) {
  ClassClass *clazz;
  int i;

  ClassClass **clazzArray = get_binclasses();
  long clazzCount = get_nbinclasses();

  for (i = clazzCount; i-- > 0; ) {
    cp_item_type *cbCP;
    cp_item_type *cbCPEnd;

    clazz = clazzArray[i];
    cbCP = cbConstantPool(clazz);
    cbCPEnd = cbCP + cbConstantPoolCount(clazz);

    if ((cp >= cbCP) && (cp < cbCPEnd))
      goto ret;
  }

  /* not found */
  return NULL;

ret:
  return clazz;
}


/*
 * ɽѿ̾롣
 */
char *LocalVarName(struct methodblock *method, int index) {
  cp_item_type *cpool = cbConstantPool(method->fb.clazz);
  struct localvar *lv = method->localvar_table + index;

  return cpool[lv->nameoff].cp;
}

/*
 * ɽѿ signature 롣
 */
char *LocalVarSig(struct methodblock *method, int index) {
  cp_item_type *cpool = cbConstantPool(method->fb.clazz);
  struct localvar *lv = method->localvar_table + index;

  return cpool[lv->sigoff].cp;
}

/*
 * Ϳ줿 PC λǶɽѿƤ뤫?
 */
int isLocalVarAlive(struct methodblock *method, int index, uint32_t pcoff) {
  cp_item_type *cpool = cbConstantPool(method->fb.clazz);
  struct localvar *lv = method->localvar_table + index;
  long start, len, end;

  start = lv->pc0;  len = lv->length;  end = start + len;

  if ((start < 0) || (len < 0))  return 0;
  
  if ((pcoff >= start) && (pcoff < end))
    return 1;
  else
    return 0;
}


/*
 *  opcode  PC 롣
 */
unsigned char *nextPC(unsigned char *pc) {
  int size_mag = 1;
  int opcode = (int)*pc;
  int opcodeLen = (int)opcode_length[opcode];
  /*
    dumpOpcodeLength(stdout);
    fflush(stdout);
   */
#ifdef DEBUG
  printf("opcode: %s(0x%x=%d), length of opcode: %d  (in nextPC())\n",
		opcode_symbol[opcode], opcode, opcode, opcodeLen);
  printf("PC: 0x%x ->", (int)pc);
#endif

  if (opcode == opc_wide) {
    opcode = *(pc+1);
    size_mag = 2;
  }

  pc += opcodeLen * size_mag;
#ifdef DEBUG
  printf(" 0x%x.\n", (int)pc);
  fflush(stdout);
#endif

  return pc;
}

/*
 * unsigned char opcode_length[]  dump 롣(for debug)
 */
void dumpOpcodeLength(FILE *out) {
  int i;
  for (i = 0; i <= opc_nonnull_quick; i++) {
    fprintf(out, "len[%d]: %d\n", i, (int)opcode_length[i]);
  }
}


/*
 * ݥ󥿤λؤ褬 handle ɤȽꤹ롣
 */
extern int isPtrHandle(void *p) {
  uint32_t flags;
  struct methodtable *methods;
  ClassClass *clazz;

#ifdef DEBUG
  printf("isPtrHandle(0x%x) called.\n", (int)p);
  fflush(stdout);
#endif
  if (!isHandle(p))  return 0;	/* false */

  if ( !validateAddress((void *)unhand((JHandle *)p)) )  return 0;

  flags = obj_flags((JHandle *)p);
  if (flags == T_NORMAL_OBJECT) {	/* ֥? */
    if ( !validateAddress((void *)(((JHandle *)p)->methods)) )  return 0;
    methods = (struct methodtable *)((JHandle *)p)->methods;

    if ( !validateAddress((void *)(methods->classdescriptor)) )  return 0;
    clazz = (ClassClass *)methods->classdescriptor;
    if ( !validateAddress((void *)cbName(clazz)) )  return 0;
    if ( !validateAddress((void *)cbSuperName(clazz)) )  return 0;
    if ( !validateAddress((void *)cbSourceName(clazz)) )  return 0;
#if 0
    printf("class name: %s\n", cbName(clazz));
    fflush(stdout);
#endif
  }
  else {	/* ? */
    if (!(flags >= T_CLASS && flags < T_MAXNUMERIC))  return 0;
  }

#ifdef DEBUG
  printf("isPtrHandle(0x%x) returns true.\n", (int)p);
  fflush(stdout);
#endif
  return 1;	/* true */
}


/*
 * PCϤ롣
 */
extern void outPC(FILE *out, unsigned char *pc, struct methodblock *method) {
  int offset;

  if (pc == 0) {
    fprintf(out, "0x0");  return;
  }

  if (!method)
    method = pc2method(pc);

  fprintf(out, "0x%x", pc);
  if (method) {
    offset = (int)pc2offset(method, pc);
    fprintf(out, " (offset: %d)", offset);
  }
  fprintf(out, " in ");
  outMethod(out, method);
}


/*
 * ᥽åɤˤĤƽϤ롣
 */
extern void outMethod(FILE *out, struct methodblock *method) {
  ClassClass *clazz;

  if (!method) {
    fprintf(out, "(null)#(null)");  fflush(out);
    return;
  }

  clazz = method->fb.clazz;

  fprintf(out, "%s#%s", (clazz ? cbName(clazz): "(null)"), method->fb.name);
  fflush(out);
}


/*
 * PC б᥽åɤ롣
 * ɤƤ뤹٤ƤΥ饹õ <- 
 */
/* following code is derived from
	javasrc/src/share/java/runtime/classruntime.c */
extern struct methodblock *pc2method(unsigned char *pc) {
  struct methodblock *method;
  ClassClass *clazz;
  int i, j;

  ClassClass **clazzArray = get_binclasses();
  long clazzCount = get_nbinclasses();
#if 0
  printf("nbinclasses: %d\n", clazzCount);  fflush(stdout);
#endif

#if (JDK_VER >= 110)
  BINCLASS_LOCK();
#endif /* JDK_VER >= 110 */
  for (i = clazzCount; i-- > 0; ) {
    clazz = clazzArray[i];
    method = cbMethods(clazz);
#if 0
    printf("class: %s\n", cbName(clazz));
    printf("nmethods: %d\n", cbMethodsCount(clazz));  fflush(stdout);
#endif
    for (j = cbMethodsCount(clazz); j-- > 0; method++) {
      if (!(method->fb.access & ACC_NATIVE)) {
	unsigned char *code_start;
	if (method->fb.access & ACC_MACHINE_COMPILED) {
	  if (PCinCompiledCode(pc, method))
	    goto ret;
	}
	else {
#if 0
	  printf(" method: %s starts 0x%x\n",
		 method->fb.name, (int)method->code);
	  fflush(stdout);
#endif
	  code_start = (unsigned char *)method->code;
	  if (pc >= code_start && pc < code_start + method->code_length)
	    goto ret;
	}
      }
    }
  }

  /* not found */
  return NULL;

ret:
#if (JDK_VER >= 110)
  BINCLASS_UNLOCK();
#endif /* JDK_VER >= 110 */
  return method;
}


/*
 * PC 򥳡ƬΥեåȤѴ롣
 * եåȤñ̤ sizeof(unsigned char)
 */
extern uint32_t pc2offset(struct methodblock *method, unsigned char *pc) {
  uint32_t ret;

  if (!pc)  ret = -1;
  else  ret = (uint32_t)(pc - method->code);

  return  ret;
}

/*
 * ƬΥեåȤ PC Ѵ롣
 */
extern unsigned char *offset2pc(struct methodblock *method, uint32_t offset) {
  if (offset == -1)
    return 0;
  else
    return (method->code + offset);
}

/*
 * åݥ󥿤򥹥å줫ΥեåȤѴ롣
 * եåȤñ̤ sizeof(stack_item)
 */
extern int16_t sp2offset(JavaFrame *frame, stack_item *sp) {
  int16_t ret;

  if (!sp)  ret = -1;
  else  ret = (int16_t)(sp - frame->ostack);

  return ret;
}

/*
 * å줫ΥեåȤ򥹥åݥ󥿤Ѵ롣
 */
extern stack_item *offset2sp(JavaFrame *frame, int16_t offset) {
  if (offset == -1)
    return 0;
  else
    return (frame->ostack + offset);
}

/*
 * 󥹥ȥס(CP)ǤؤΥݥ󥿤򥪥եåȤѴ롣
 */
extern uint16_t cp2offset(ClassClass *clazz, cp_item_type *cp_item) {
  uint16_t ret;

  if (!cp_item)  ret = -1;
  else  ret = (uint16_t)(cp_item - cbConstantPool(clazz));

  return ret;
}

/*
 * եåȤ CP ǤؤΥݥ󥿤Ѵ롣
 */
extern cp_item_type *offset2cp(ClassClass *clazz, uint16_t offset) {
  if ((int16_t)offset == -1)
    return 0;
  else
    return (cbConstantPool(clazz) + offset);
}


/*
 * ᥽åɤƤӽФ
 */
extern void *callMethod(JNIEnv *env, jobject obj, jmethodID mid, int staticCall, ...) {
  ExecEnv *ee = JNIEnv2EE(env);
  JHandle *handle;
  struct methodblock *method = (struct methodblock *)mid;
  bool_t isStaticCall = (staticCall ? TRUE:FALSE);
  va_list origargs, args;
  char *sig;
  long ret;

#ifdef DEBUG
  printf("callMethod() called.\n");
  fflush(stdout);
#endif
  handle = (obj ? DeRef(env, obj) : NULL);

  sig = method->fb.signature;
  if (*sig != '(') {
#ifdef DEBUG
    printf("callMethod(): illegal methodID.\n");  fflush(stdout);
    return NULL;
#endif
  }
  sig++;

  /*  (jobject) -> (JHandle *) Ѵ */
  va_start(origargs, staticCall);
  va_start(args, staticCall);
  while (*sig != ')') {
    jobject anArg;

    switch (*sig) {
    case '[':
      anArg = va_arg(origargs, jobject);
      va_arg(args, JHandle *) = DeRef(env, anArg);
      sig++;
      if (*sig == 'L')  while (*sig != ';')  sig++;
      sig++;
      break;
    case 'L':
      anArg = va_arg(origargs, jobject);
      va_arg(args, JHandle *) = DeRef(env, anArg);
      sig++;
      while (*sig != ';')  sig++;
      sig++;
      break;
    default:
      va_arg(origargs, void *);
      va_arg(args, void *);
      break;
    }
  }
  va_end(origargs);
  va_end(args);

#ifdef DEBUG
  printf("call do_execute_java_method_vararg(0x%x, 0x%x, , , 0x%x, ...).\n",
		(int)ee, (int)handle, (int)method);
  {
    ClassClass *cb = handle->methods->classdescriptor;
    printf(" class: %s\n", cbName(cb));
    printf(" method: %s, %s\n", method->fb.name, method->fb.signature);
  }
  fflush(stdout);
#endif

  /* ᥽åɸƤӽФ */
  va_start(args, staticCall);
  ret = do_execute_java_method_vararg(ee,
		/* object */ handle,
		/* method name */ NULL,
		/* method signature */ NULL,
		/* method block of */ (struct methodblock *)method,
		/* isStaticCall */ isStaticCall,
		/* argument */ args,
			0, FALSE);
  va_end(args);
#ifdef DEBUG
  printf("do_execute_java_method_vararg() done.\n");
  fflush(stdout);
#endif

#ifdef DEBUG
  printf("callMethod() returns: 0x%x\n", (int)ret);
  fflush(stdout);
#endif
  return (void *)ret;
}


/*
 * 饹饯饹롣
 */
extern jclass classLoaderFindClass(JNIEnv *env, char *name, jobject loader) {
  ExecEnv *ee = JNIEnv2EE(env);
  ClassClass *cb;

  Classjava_lang_Class *from;
  ClassClass *fromHandle;

#ifdef DEBUG
  printf("classLoaderFindClass() called.\n");
  fflush(stdout);
#endif
  from = (Classjava_lang_Class *)malloc(sizeof(Classjava_lang_Class));
  fromHandle = (ClassClass *)malloc(sizeof(ClassClass));
  unhand(fromHandle) = from;
  cbLoader(fromHandle) = DeRef(env, loader);
  cb = FindClassFromClass(ee, name, TRUE, fromHandle);
#ifdef DEBUG
  printf("FindClassFromClass() returns : 0x%x\n", (int)cb);
  fflush(stdout);
#endif
  free(from);
  free(fromHandle);

  if (cb)
    return MkRefLocal(env, cb, JNI_REF_HANDLE_TAG);
  else
    return NULL;
}


/*
 * htonl, htons, ntohl, ntohs  64bit 
 */
#ifndef BIG_ENDIAN
extern jlong hton64(jlong x) {
  jlong ret;

  *(unsigned long *)&ret = htonl((unsigned long)(x >> 32));
  *(((unsigned long *)&ret) + 1) = htonl((unsigned long)x);

  return ret;
}

extern jlong ntoh64(jlong x) {
  jlong ret = 0;

  ret |= ( ntohl(*(unsigned long *)&x) | 0LL) << 32;
  ret |= ntohl(*(((unsigned long *)&x) + 1)) | 0LL;

  return ret;
}
#endif	/* BIG_ENDIAN */
